from __future__ import annotations
from typing import Optional, Union
from sympy.physics.wigner import wigner_6j, wigner_9j
from sympy import sqrt
from math import prod
import logging

from stolpy.tensor_operator import TensorOperator, TensorOperatorComposite

logging.basicConfig(level=logging.WARNING)


def jsc(
    *args,
):
    """
    Auxiliary function for a reappearing element during recoupling actions [j], where [j] = sqrt(2 * j + 1)

    :param args: either a single integer/Symbol or multiple integers/Symbols
    :return: the product of the input
    """
    return prod([sqrt(2 * arg + 1) for arg in args])


class TensorAlgebra(object):
    """
    Naming convention for tensor operator recoupling:
    The recoupling operations work for instance as follows (A x B) x (C x D) -> (A x C) x (B x D). We define our naming
    convention according to the initial space and final result. The functions all start with _recouple_initial_final.
    The example from above would require the function
    _recouple_ABxCD_ACxBD
    """

    @staticmethod
    def _can_be_recoupled_ABxAB_AAxBB(
        tensor_op: TensorOperator,
    ) -> bool:
        """
        Checks the subspaces of this operator. If the operator subspace structure returns two mixed states, with space
        one = A and space two = B coupled like (AxB) x (AxB) we can recouple it to (AxA) x (BxB) and the function
        returns True, otherwise it returns False.

        :param tensor_op: A TensorOperator for which we want to check whether it can be recoupled.
        :return: bool True or False. True -> Can be recoupled. False -> Cannot be recoupled.
        """
        if not tensor_op.substructure:
            return False
        first_pair, second_pair = tensor_op.substructure
        if not (first_pair.substructure and second_pair.substructure):
            return False
        tensor_a, tensor_b = first_pair.substructure
        tensor_c, tensor_d = second_pair.substructure
        if tensor_a.space == tensor_c.space and tensor_b.space == tensor_d.space and tensor_a.space != tensor_b.space:
            return True
        return False

    @staticmethod
    def _can_be_recoupled_AAxAB_AAAxB(
        tensor_op: TensorOperator,
    ) -> bool:
        """
        Checks the subspaces of this operator. If the operator subspace structure returns two mixed states, with space
        one = A and space two = B coupled like (AxA) x (AxB) we can recouple it to ((AxA) x A) x B and the function
        returns True, otherwise it returns False.

        :param tensor_op: A TensorOperator for which we want to check whether it can be recoupled.
        :return: bool True or False. True -> Can be recoupled. False -> Cannot be recoupled.
        """
        if not tensor_op.substructure:
            return False
        first_pair, second_pair = tensor_op.substructure
        if not (first_pair.substructure and second_pair.substructure):
            return False
        tensor_a, tensor_b = first_pair.substructure
        tensor_c, tensor_d = second_pair.substructure
        if (
            tensor_a.space <= tensor_b.space
            and tensor_a.space <= tensor_c.space
            and max([tensor_a.space.order, tensor_b.space.order, tensor_c.space.order]) < tensor_d.space.order
        ):
            return True
        return False

    @staticmethod
    def _can_be_recoupled_ABxC_ACxB(
        tensor_op: TensorOperator,
    ) -> bool:
        """
        Checks the subspaces of this operator. If the operator subspace structure returns one mixed state and one pure
        state, with space one = AxB and space two = C coupled like (AxB) x C we can recouple it to (AxC) x B given that
        either space A = space C != space B or order space A >= order space C > order space B.

        :param tensor_op: A TensorOperator for which we want to check whether it can be recoupled.
        :return: bool True or False. True -> Can be recoupled. False -> Cannot be recoupled.
        """
        if not tensor_op.substructure:
            return False
        first_pair, tensor_c = tensor_op.substructure
        if not first_pair.substructure:
            return False
        tensor_a, tensor_b = first_pair.substructure
        if (
            tensor_a.space == tensor_b.space
            or tensor_b.space == tensor_c.space
            or tensor_b.space.contains(tensor_c.space)
        ):
            return False
        elif tensor_a.space == tensor_c.space or tensor_a.space <= tensor_c.space < tensor_b.space:
            return True
        return False

    @staticmethod
    def _can_be_recoupled_ABxC_AxBC(
        tensor_op: TensorOperator,
    ) -> bool:
        """
        Checks the subspaces of this operator. If the operator subspace structure returns one mixed state and one pure
        state, with space one = AxB and space two = C coupled like (AxB) x C we can recouple it to A x (BxC) given that
        either space A != space C and space C == space B.

        :param tensor_op: A TensorOperator for which we want to check whether it can be recoupled.
        :return: bool True or False. True -> Can be recoupled. False -> Cannot be recoupled.
        """
        if not tensor_op.substructure:
            return False
        first_pair, tensor_c = tensor_op.substructure
        if not first_pair.substructure:
            return False
        tensor_a, tensor_b = first_pair.substructure
        if tensor_a.space == tensor_b.space or tensor_a.space == tensor_c.space:
            return False
        elif tensor_a.space.order < min([tensor_c.space.order, tensor_b.space.order]) and (
            tensor_b.space.contains(tensor_c.space) or tensor_c.space.contains(tensor_b.space)
        ):
            return True
        return False

    @staticmethod
    def _can_be_recoupled_AxBC_ABxC(
        tensor_op: TensorOperator,
    ) -> bool:
        """
        Checks the subspaces of this operator. If the operator subspace structure returns one mixed state and one pure
        state, with space one = A and space two = BxC coupled like A x (BxC) we can recouple it to (AxB) x C given that
        either space B != space C and space A == space B.

        :param tensor_op: A TensorOperator for which we want to check whether it can be recoupled.
        :return: bool True or False. True -> Can be recoupled. False -> Cannot be recoupled.
        """
        if not tensor_op.substructure:
            return False
        tensor_a, second_pair = tensor_op.substructure
        if not second_pair.substructure:
            return False
        tensor_b, tensor_c = second_pair.substructure
        if tensor_b.space == tensor_c.space or tensor_a.space == tensor_c.space:
            return False
        elif tensor_c.space.order >= max([tensor_a.space.order, tensor_b.space.order]) and (
            tensor_a.space.contains(tensor_b.space) or tensor_b.space.contains(tensor_a.space)
        ):
            return True
        return False

    @classmethod
    def _perform_recoupling(
        cls,
        tensor_op: TensorOperator,
        factor=1,
    ) -> Union[None, TensorOperator, TensorOperatorComposite]:
        """
        First check whether any of the implemented recoupling methods can be applied. If so run the recoupling routine
        and return the output. If not return the original tensor.

        :param tensor_op: TensorOperator
        :param factor: optional factor that can be applied during the recoupling
        :return: TensorOperator or TensorOperatorComposite
        """
        out_tensor = tensor_op
        if cls._can_be_recoupled_ABxAB_AAxBB(out_tensor):
            out_tensor = cls._recouple_ABxCD_ACxBD(out_tensor, factor)
        elif cls._can_be_recoupled_AAxAB_AAAxB(out_tensor):
            out_tensor = cls._recouple_ABxCD_ABCxD(out_tensor, factor)
        elif cls._can_be_recoupled_ABxC_ACxB(out_tensor):
            out_tensor = cls._recouple_ABxC_ACxB(out_tensor, factor)
        elif cls._can_be_recoupled_ABxC_AxBC(out_tensor):
            out_tensor = cls._recouple_ABxC_AxBC(out_tensor, factor)
        return out_tensor

    @classmethod
    def _perform_recoupling_composite(
        cls,
        tensor_op: TensorOperatorComposite,
        factor=1,
    ) -> Union[None, TensorOperator, TensorOperatorComposite]:
        """
        See also _perform_recoupling method. This method is called for composites and loops through the children.

        :param tensor_op: TensorOperatorComposite
        :param factor: An optional factor that can be applied during the recoupling
        :return: A recoupled TensorOperator or TensorOperatorComposite
        """
        out_tensor = cls._perform_recoupling(tensor_op.children[0], factor)
        for child in tensor_op.children[1:]:
            res = cls._perform_recoupling(child, factor)
            if not res:
                continue
            out_tensor += res

        return out_tensor

    @classmethod
    def _recouple_substructure(
        cls,
        tensor_op: Union[TensorOperator, TensorOperatorComposite],
    ) -> Union[None, TensorOperator, TensorOperatorComposite]:
        """
        Performs an incremental recoupling step (see also _recouple_basic_substructure). If the input is a
        TensorOperatorComposite, it loops through all children and performs this step.

        :param tensor_op: TensorOperator or TensorOperatorComposite
        :return: TensorOperator or TensorOperatorComposite
        """
        if isinstance(tensor_op, TensorOperator):
            return cls._recouple_basic_substructure(tensor_op)
        elif isinstance(tensor_op, TensorOperatorComposite):
            out_tensor = cls._recouple_basic_substructure(tensor_op.children[0])
            for child in tensor_op.children[1:]:
                res = cls._recouple_basic_substructure(child)
                if not res:
                    continue
                out_tensor += res
            return out_tensor

    @classmethod
    def _recouple_basic_substructure(
        cls,
        tensor_op: Union[TensorOperator, TensorOperatorComposite],
    ) -> Union[None, TensorOperator, TensorOperatorComposite]:
        """
        Idea: Recoupling should take place starting from the innermost operators going to the outermost operators.
        A recursive call to the recouple method takes care of this.

        :param tensor_op: An operator of type TensorOperator or TensorOperatorComposite
        :return: None, TensorOperator, or TensorOperatorComposite
        """
        out_tensor = tensor_op
        if out_tensor.get_depth() > 2:  # The tensor operator has a substructure that can potentially be recoupled
            sub_tensor_1, sub_tensor_2 = out_tensor.substructure
            new_sub_tensor_1 = cls.recouple(sub_tensor_1, outer_loop=False)
            new_sub_tensor_2 = cls.recouple(sub_tensor_2, outer_loop=False)
            if new_sub_tensor_1 != sub_tensor_1 or new_sub_tensor_2 != sub_tensor_2:
                out_tensor = new_sub_tensor_1.couple(new_sub_tensor_2, out_tensor.rank, out_tensor.factor)
        return out_tensor

    @classmethod
    def _recouple_step(
        cls,
        tensor_op: Union[TensorOperator, TensorOperatorComposite],
        factor=1,
    ) -> Optional[TensorOperator, TensorOperatorComposite]:
        """
        Basic recouple step for a given tensor_op. Depending on whether the input is a TensorOperator or
        TensorOperatorComposite, the methods _perform_recoupling or _perform_recoupling_composite are called.
        See also their description.

        :param tensor_op: TensorOperator or TensorOperatorComposite
        :param factor: optional factor that can be applied during the recoupling
        :return: TensorOperator, TensorOperatorComposite, or None
        """
        out_tensor = cls._recouple_substructure(tensor_op)  # First recouple innermost layers

        if isinstance(out_tensor, TensorOperator):
            out_tensor = cls._perform_recoupling(out_tensor, factor)
        elif isinstance(out_tensor, TensorOperatorComposite):
            out_tensor = cls._perform_recoupling_composite(out_tensor, factor)
        return out_tensor

    @classmethod
    def recouple(
        cls,
        tensor_op: Union[TensorOperator, TensorOperatorComposite],
        factor=1,
        outer_loop=True,
    ) -> Union[TensorOperator, TensorOperatorComposite]:
        """
        Perform the action recoupling if possible. The recoupling takes place in the _recouple_step method. When the
        outer_loop flag is set to True this method keeps track of changes and repeats the recoupling until no further
        changes are detected. During the repetition, the outer_loop flag is set to false.

        :param tensor_op: The TensorOperator and TensorOperatorComposite under investigation
        :param factor: an optional factor that can be applied during the recoupling
        :param outer_loop: bool
        :return: TensorOperator or TensorOperatorComposite
        """
        out_tensor = cls._recouple_step(tensor_op, factor)

        if out_tensor == tensor_op:
            logging.info(f"Recoupling was not possible for tensor operator {tensor_op}")
        if out_tensor != tensor_op and outer_loop:
            prev_tensor = None
            while prev_tensor != out_tensor:  # Repeat recoupling steps until no further change happens
                prev_tensor = out_tensor
                out_tensor = cls.recouple(out_tensor, factor, outer_loop=False)
        return out_tensor

    @staticmethod
    def _recouple_ABxCD_ACxBD(
        tensor_op: TensorOperator,
        factor=1,
    ) -> Optional[TensorOperator]:
        """
        Tensor recoupling of the form
        (A_a x B_b)_ab x (C_c x D_d)_cd -> (A_a x C_c)_ac x (B_b x D_d)_bd
        The assumption is, that these operators are already ordered according to their spaces within the brackets and
        that space A = space C and space B = space D.

        :param tensor_op: the tensor operator that should be recoupled
        :param factor: a factor that appears during the recoupling, default is 1
        :return: recoupled tensor operator of type TensorOperator or TensorOperatorComposite
        """
        first_pair, second_pair = tensor_op.substructure
        tensor_a, tensor_b = first_pair.substructure
        tensor_c, tensor_d = second_pair.substructure
        new_factor = tensor_op.factor * factor
        a = tensor_a.rank
        b = tensor_b.rank
        c = tensor_c.rank
        d = tensor_d.rank
        ab = first_pair.rank
        cd = second_pair.rank
        rank = tensor_op.rank

        out_tensor = None
        for ac in range(abs(a - c), a + c + 1):
            for bd in range(abs(b - d), b + d + 1):
                if not (abs(ac - bd) <= rank <= ac + bd):
                    continue
                operator_factor = new_factor * jsc(ab, cd, ac, bd) * wigner_9j(a, b, ab, c, d, cd, ac, bd, rank)
                if operator_factor != 0:
                    new_first_pair = tensor_a.couple(tensor_c, ac, 1)
                    new_second_pair = tensor_b.couple(tensor_d, bd, 1)
                    if not new_first_pair or not new_second_pair:
                        continue
                    new_tensor = new_first_pair.couple(new_second_pair, rank, operator_factor)
                    if not out_tensor:
                        out_tensor = new_tensor
                    else:
                        out_tensor += new_tensor
        logging.info("recouple ABxCD -> ACxBD")
        logging.debug(f"{tensor_op} -> {out_tensor}")
        return out_tensor

    @staticmethod
    def _recouple_ABxCD_ABCxD(
        tensor_op: TensorOperator,
        factor=1,
    ) -> Optional[TensorOperator]:
        """
        Tensor recoupling of the form
        (A_a x B_b)_ab x (C_c x D_d)_cd -> ((A_a x B_b)_ab x C_c)_abc x D_d
        The assumption is, that these operators are already ordered according to their spaces within the brackets and
        that space A = space B and space C != space D

        :param tensor_op: the tensor operator that should be recoupled
        :param factor: a factor that appears during the recoupling, default is 1
        :return: recoupled tensor operator of type TensorOperator or TensorOperatorComposite
        """
        first_pair, second_pair = tensor_op.substructure
        tensor_c, tensor_d = second_pair.substructure
        new_factor = tensor_op.factor * factor
        c = tensor_c.rank
        d = tensor_d.rank
        ab = first_pair.rank
        cd = second_pair.rank
        rank = tensor_op.rank
        out_tensor = None
        for abc in range(abs(c - ab), c + ab + 1):
            operator_factor = (
                jsc(cd, abc) * wigner_6j(c, ab, abc, rank, d, cd) * new_factor * pow(-1, d + c + ab + rank)
            )
            if operator_factor != 0:
                triplet = first_pair.couple(tensor_c, abc, 1)
                if not triplet:
                    continue
                new_tensor = triplet.couple(tensor_d, rank, operator_factor)
                if not out_tensor:
                    out_tensor = new_tensor
                else:
                    out_tensor += new_tensor
        logging.info("recouple ABxCD -> ABCxD")
        logging.debug(f"{tensor_op} -> {out_tensor}")
        return out_tensor

    @staticmethod
    def _recouple_ABxC_ACxB(
        tensor_op: TensorOperator,
        factor=1,
    ) -> Optional[TensorOperator]:
        """
        Tensor recoupling of the form
        (A_a x B_b)_ab x C_c -> (A_a x C_c)_ac x B_b
        The assumption is, that these operators are already ordered according to their spaces within the brackets.
        This recoupling scheme is applied when the order of space B is higher than that of space C and
        space A != space B.

        :param tensor_op: the tensor operator that should be recoupled
        :param factor: a factor that appears during the recoupling, default is 1
        :return: recoupled tensor operator of type TensorOperator or TensorOperatorComposite
        """
        first_pair, tensor_c = tensor_op.substructure
        tensor_a, tensor_b = first_pair.substructure
        new_factor = tensor_op.factor * factor
        a = tensor_a.rank
        b = tensor_b.rank
        c = tensor_c.rank
        ab = first_pair.rank
        rank = tensor_op.rank
        out_tensor = None
        for ac in range(abs(a - c), a + c + 1):
            operator_factor = jsc(ab, ac) * wigner_6j(a, b, ab, rank, c, ac) * new_factor * pow(-1, ac + c)
            if operator_factor != 0:
                new_pair = tensor_a.couple(tensor_c, ac, 1)
                if not new_pair:
                    continue
                new_tensor = new_pair.couple(tensor_b, rank, operator_factor)
                if not out_tensor:
                    out_tensor = new_tensor
                else:
                    out_tensor += new_tensor
        logging.info("recouple ABxC -> ACxB")
        logging.debug(f"{tensor_op} -> {out_tensor}")
        return out_tensor

    @staticmethod
    def _recouple_ABxC_AxBC(
        tensor_op: TensorOperator,
        factor=1,
    ) -> Optional[TensorOperator]:
        """
        Tensor recoupling of the form
        (A_a x B_b)_ab x C_c -> A_a x (B_b x C_c)_bc
        The assumption is, that these operators are already ordered according to their spaces within the brackets.
        This recoupling scheme is applied when space B = space C and space A != space B.

        :param tensor_op: the tensor operator that should be recoupled
        :param factor: a factor that appears during the recoupling, default is 1
        :return: recoupled tensor operator of type TensorOperator or TensorOperatorComposite
        """

        first_pair, tensor_c = tensor_op.substructure
        tensor_a, tensor_b = first_pair.substructure
        a = tensor_a.rank
        b = tensor_b.rank
        c = tensor_c.rank
        ab = first_pair.rank
        rank = tensor_op.rank
        out_tensor = None
        new_factor = tensor_op.factor * factor * pow(-1, b + c - ab - rank)
        for bc in range(abs(b - c), b + c + 1):
            operator_factor = jsc(ab, bc) * wigner_6j(b, a, ab, rank, c, bc) * new_factor
            if operator_factor != 0:
                new_pair = tensor_b.couple(tensor_c, bc, 1)
                if not new_pair:
                    continue
                new_tensor = tensor_a.couple(new_pair, rank, operator_factor)
                if not out_tensor:
                    out_tensor = new_tensor
                else:
                    out_tensor += new_tensor
        logging.info("recouple ABxC -> AxBC")
        logging.debug(f"{tensor_op} -> {out_tensor}")
        return out_tensor



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Union, Optional
from copy import deepcopy

from stolpy.tensor_space import TensorSpace

default_space = TensorSpace("DEFAULT", 99)


class TensorOperatorInterface(ABC):
    """
    Interface for the TensorOperator class and the TensorOperatorComposite class
    """

    def __add__(
        self,
        other,
    ):
        return self.add(other)

    def __radd__(
        self,
        other,
    ):
        return self.__add__(other)

    def __sub__(
        self,
        other,
    ):
        return self.add(-1 * other)

    def __str__(
        self,
    ):
        return self.to_expression()

    def __repr__(
        self,
    ):
        return self.__str__()

    @abstractmethod
    def __mul__(
        self,
        factor,
    ):
        """
        Multiply a float/integer/symbol/etc. to the factor of this tensor operator / composite

        :param factor: float/integer/symbol/etc.
        :return: A new tensor operator / composite with the new factor = existing factor * factor
        """
        pass

    def __rmul__(
        self,
        factor,
    ):
        return self.__mul__(factor)

    def __truediv__(
        self,
        other,
    ):
        return self * (1 / other)

    def __rtruediv__(
        self,
        other,
    ):
        return self.__truediv__(other)

    @abstractmethod
    def to_expression(
        self,
    ) -> str:
        """
        Return a string expression for the given object. The expression for a Tensor Operator should show the factor
        times an expression for the tensor operator, e.g., 15 * {A_a x B_b}_c. For composites, it should return the
        string expression of all children, connected with a + sign.

        :return: String
        """
        pass

    @abstractmethod
    def to_expression_no_factor(
        self,
    ) -> str:
        """
        Return a string expression similar to to_expression, but without explicitly printing the factor.

        :return: String
        """
        pass

    @abstractmethod
    def add(
        self,
        other: TensorOperatorInterface,
    ) -> TensorOperatorInterface:
        """
        Add two TensorOperators/Composites.
        In case we add two TensorOperators -> TensorOperator with new factor if the operators are identical up to the
        factor. The new factor is just the sum of the old factors. Otherwise, return a TensorOperatorComposite with the
        two TensorOperators as Children. In all other Cases (e.g. adding Two TensorOperatorComposites) return a new
        composite with an updated list of children.

        :param other: TensorOperator or TensorOperatorComposite
        :return: TensorOperator or TensorOperatorComposite
        """
        pass

    @abstractmethod
    def couple(
        self,
        other: TensorOperatorInterface,
        rank,
        factor,
    ) -> TensorOperatorInterface:
        """
        Allows to couple TensorOperators and Composites with each other.
        E.g., A_a.couple(B_b, c, 15) -> 15 * {A_a x B_b}_c. Similar one can couple each instance in a Composite to an
        Operator or another Composite.

        :param other: TensorOperator or TensorOperatorComposite
        :param rank: The new rank to which the operators are coupled, checks for triangular equations
        :param factor: A new factor that is multiplied to existing factors
        :return: TensorOperator or TensorOperatorComposite depending on the input or None if triangular equations are
        violated
        """
        pass


class TensorOperatorComposite(TensorOperatorInterface):
    """
    This class allows to create objects that hold a sum of TensorOperators, and still allows to perform calculations
    with its children.

    The class roughly follows a Composite Pattern, where this class is the composite and TensorOperators are the leafs.
    However, this class is flat. It cannot hold instances of itself.
    """

    def __init__(
        self,
        *args,
    ):
        self._children = [arg for arg in args if arg]
        self._simplify()

    @property
    def children(
        self,
    ):
        return self._children

    @children.setter
    def children(
        self,
        children,
    ):
        self._children = children
        self._simplify()

    def __mul__(
        self,
        factor,
    ):
        """
        Multiply each of the children with a factor

        :param factor: integer/float/symbol
        :return: A new instance of this class with an updated list of children
        """
        args = []
        for child in self.children:
            args.append(child * factor)
        return self.__class__(*args)

    def __eq__(
        self,
        other: TensorOperatorInterface,
    ) -> bool:
        if not isinstance(other, self.__class__):
            return False
        if len(self.children) != len(other.children):
            return False
        return all([child1 == child2 for child1, child2 in zip(self.children, other.children)])

    def __ne__(
        self,
        other: TensorOperatorInterface,
    ) -> bool:
        return not self.__eq__(other)

    def to_expression(
        self,
    ) -> str:
        """
        Returns a string expression for the sum of all children.
        :return: String
        """
        return " + ".join([child.to_expression() for child in self.children])

    def to_expression_no_factor(
        self,
    ) -> str:
        """
        Returns a string expression for the sum of all children without explicitly showing the factors.
        :return: String
        """
        return " + ".join([child.to_expression_no_factor() for child in self.children])

    def add(
        self,
        other: TensorOperatorInterface,
    ) -> TensorOperatorComposite:
        """
        Extends the list of children depending on the input and returns a new instance of this class with the updated
        list.

        :param other: TensorOperator or TensorOperatorComposite
        :return: TensorOperatorComposite
        """
        if isinstance(other, self.__class__):
            args = self.children + other.children  # This keeps a flat hierarchy
        else:
            args = self.children + [other]
        new_object = self.__class__(*args)
        return new_object

    def couple(
        self,
        other: TensorOperatorInterface,
        rank,
        factor,
    ) -> TensorOperatorComposite:
        """
        Couple a TensorOperator or TensorOperatorComposite to all children of this object and return a new instance
        of this class with an updated list of children.

        :param other: TensorOperator or TensorOperatorComposite
        :param rank: The new rank to which the tensor operator should be coupled
        :param factor: A factor that is applied during the recoupling
        :return: A new TensorOperatorComposite
        """
        if isinstance(other, self.__class__):
            args = [
                child.couple(other_child, rank, factor) for child in self.children for other_child in other.children
            ]
        else:
            args = [child.couple(other, rank, factor) for child in self.children]

        new_object = self.__class__(*args)
        return new_object

    def _simplify(
        self,
    ):
        """
        Auxiliary function that simplifies the list of children in place.
        It removes zero entries and combines matching entries to one.
        This action is called whenever the children setter is used or a new instance of this object is created.

        :return: None
        """
        children = self.children
        new_children = []
        while len(children) > 0:
            child = children.pop(0)
            if not child:  # Remove bad couplings
                continue
            for i, other_child in enumerate(children):
                if child == other_child:
                    children.pop(i)
                    child += other_child
            if child:
                new_children.append(child)
        self._children = new_children


class TensorOperator(TensorOperatorInterface):
    """
    This class holds basic (no substructure) and coupled components (substructure) of tensor operators.

    The class roughly follows a Composite Pattern, where the TensorOperatorComposite class is the composite and
    this class are the leafs.
    """

    def __init__(
        self,
        rank,
        factor=1,
        space: TensorSpace = default_space,
        symbol: Union[None, str, list[str]] = None,
        substructure=None,
    ):
        self._rank = rank
        self._factor = factor
        self._space = space
        self._symbol = symbol
        self._substructure = substructure

    def __mul__(
        self,
        factor,
    ) -> Optional[TensorOperator]:
        """
        Allows to multiply a factor to the tensor operator. In case the factor is 0, the method returns one. Otherwise
        a new instance of this class with an updated factor is returned.

        :param factor: inst/float/symbol
        :return: TensorOperator or None
        """
        new_factor = self.factor * factor
        if new_factor != 0:
            return self.__class__(self.rank, new_factor, self.space, self.symbol, self.substructure)
        else:
            return None

    def __eq__(
        self,
        other: Union[TensorOperatorComposite, TensorOperator],
    ) -> bool:
        if not isinstance(other, self.__class__):
            return False
        elif self.space == other.space and self.symbol == other.symbol and self.rank == other.rank:
            return True
        else:
            return False

    def __ne__(
        self,
        other,
    ) -> bool:
        return not self.__eq__(other)

    @property
    def rank(
        self,
    ):
        return self._rank

    @property
    def factor(
        self,
    ):
        return self._factor

    @factor.setter
    def factor(
        self,
        value,
    ):
        self._factor = value

    @property
    def symbol(
        self,
    ):
        return self._symbol

    @symbol.setter
    def symbol(self, value):
        self._symbol = value

    @property
    def space(
        self,
    ):
        return self._space

    @space.setter
    def space(
        self,
        value,
    ):
        self._space = value

    @property
    def substructure(
        self,
    ):
        return self._substructure

    @substructure.setter
    def substructure(
        self,
        value,
    ):
        self._substructure = value

    def get_depth(
        self,
    ) -> int:
        """
        Auxiliary function to get the maximum coupling depth of the operator. This is done by going through the
        substructure recursively.

        :return: Integer value
        """
        if not self.substructure:
            return 0
        else:
            sub_structure_depth = max([sub_tensor.get_depth() for sub_tensor in self.substructure]) + 1
            return sub_structure_depth

    def add(
        self,
        other: Union[TensorOperator, TensorOperatorComposite],
    ) -> Union[TensorOperator, TensorOperatorComposite, None]:
        """
        Allows the addition of two tensor operators. If the tensor operators are identical up to the factor, return an
        updated instance of this object with the new factor, which is the sum of both factors. If this new factor is
        zero, return None. In all other cases, the method returns a TensorOperatorComposite.

        :param other: TensorOperator or TensorOperatorComposite
        :return: TensorOperator or TensorOperatorComposite or None
        """
        if isinstance(other, TensorOperatorComposite):
            return other + self
        elif isinstance(other, TensorOperator) and self.symbol == other.symbol and self.rank == other.rank:
            new_factor = self.factor + other.factor
            if new_factor != 0:
                return self.__class__(self.rank, new_factor, self.space, self.symbol, self.substructure)
            else:
                return None
        else:
            return TensorOperatorComposite(self, other)

    def to_expression(
        self,
    ) -> str:
        """
        A string expression of the operator at hand.

        :return: String
        """
        return str(self.factor) + " * " + self.to_expression_no_factor()

    def to_expression_no_factor(
        self,
    ) -> str:
        """
        A string expression of the operator at hand, without explicitly listing the factor.

        :return: String
        """
        return (
            str(self.symbol).replace("[", "{").replace("]", "}").replace(", ", " x ").replace("'", "")
            + "_"
            + str(self.rank)
        )

    def couple(
        self,
        other: TensorOperatorInterface,
        rank,
        factor,
        order=True,
    ) -> Optional[TensorOperatorInterface]:
        """
        Couple two Tensor Operators together to a new Tensor Operator of a given rank.
        There are two simplifications. First, the new rank has to fulfill
        |self.rank - other.rank| <= rank <= self.rank + other.rank
        Furthermore, Tensor Operators of rank 1 are vectors and the tensor product of two vectors to rank 1 is
        basically a vector product. The tensor product of two parallel vectors is zero. This means, e.g.,
        {q_1 x q_1}_1 = 0. We set such terms to zero.

        The new operator keeps track of its constituents in the self._substructure variable, all factors are carried
        to the outermost level.

        :param other: Either TensorOperator or TensorOperatorComposite
        :param rank:  The new rank to which the two entities shall be coupled to
        :param factor: Any factors that need to be added to the coupling
        :param order: default True, flag that indicates whether the coupled result shall be ordered
        :return: A new, coupled instance of TensorOperator or TensorOperatorComposite (depending on "other")
        """
        if isinstance(other, TensorOperatorComposite):
            args = [self.couple(other_child, rank, factor) for other_child in other.children]
            return TensorOperatorComposite(*args)
        elif isinstance(other, self.__class__):
            # Coupling two operators to the new rank is not possible
            if not abs(self.rank - other.rank) <= rank <= self.rank + other.rank:
                return None

            # Coupling two identical operators to rank 1 is a cross product, i.e., zero for parallel vectors
            if self == other and rank == 1:
                return None

            # Valid Cases
            new_factor = self.factor * other.factor * factor
            if self.space == other.space:
                new_space = self.space
            else:
                new_space = self.space + other.space
            new_symbol = [self.to_expression_no_factor(), other.to_expression_no_factor()]
            # keep track of factors on outermost layer
            tensor_a = deepcopy(self)
            tensor_b = deepcopy(other)
            tensor_a.factor = 1
            tensor_b.factor = 1
            substructure = [tensor_a, tensor_b]
            new_object = self.__class__(rank, new_factor, new_space, new_symbol, substructure)
            if order:
                new_object.order()
            return new_object

    def commute(
        self,
    ) -> Optional[TensorOperator]:
        """
        In case this tensor operator has a substructure, it commutes the substructure and applies the correct factor to
        it. A new instance of this class is returned.
        :return: TensorOperator or None
        """
        if self.substructure:
            tensor_a, tensor_b = self.substructure
            new_factor = pow(-1, tensor_a.rank + tensor_b.rank - self.rank)
            new_top = tensor_b.couple(tensor_a, self.rank, new_factor * self.factor, order=False)
            return new_top
        else:
            return None

    def order(
        self,
    ):
        """
        Order the tensor operators according to
            1st) their space (alphabetically)
            2nd) their structure depth (greater depth first), e.g., {q_1 x q_1}_2 x k_1 should stay that way
            3rd) their symbol (alphabetically)
        The idea is to have a consistent representation. The order itself does not matter.

        :return: None
        """
        new_top = None
        if self.substructure:  # ordering only makes sense if there is a substructure
            tensor_a, tensor_b = self.substructure
            commute = False

            if tensor_a.space > tensor_b.space:
                commute = True
            elif tensor_a.space == tensor_b.space:
                if tensor_b.get_depth() > tensor_a.get_depth():  # depth
                    commute = True
                elif tensor_b.get_depth() == tensor_a.get_depth() and str(tensor_a.symbol) > str(
                    tensor_b.symbol
                ):  # alphabetically
                    commute = True

            if commute:
                new_top = self.commute()

        if new_top:
            self.factor = new_top.factor
            self.symbol = new_top.symbol
            self.space = new_top.space
            self.substructure = new_top.substructure

    def get_space_structure(
        self,
    ):
        """
        Goes recursively through all operators and obtains the space object of each operator. The function returns a
        (nested) List with all spaces.

        :return: (nested) list containing all spaces for each substructure operator.
        """
        depth = self.get_depth()
        if depth == 0:
            return self.space
        elif depth == 1:
            tensor_a, tensor_b = self.substructure
            return [tensor_a.space, tensor_b.space]
        else:
            tensors_a, tensors_b = self.substructure
            return [tensors_a.get_space_structure(), tensors_b.get_space_structure()]



from __future__ import annotations
from collections.abc import Iterable, Iterator
from typing import Optional, List


class DuplicateSpaceError(Exception):
    """
    Thrown when a name for a space is already taken, but the user tries to define a new space with the same name and
    different properties.
    """


class SpaceIterator(Iterator):
    """
    Iterator to pass through Space objects and their subspaces.
    Following the example from https://refactoring.guru/design-patterns/iterator/python/example

    We want to have a standardized way to iterate through space couplings. Here is one example how we want the traversal
    to go for three spaces, labeled 1, 2, and 3
    {{1 x 2} x {1 x 3}}
    0. {{1 x 2} x {1 x 3}}
    1. {1 x 2}
    2. 1
    3. 2
    4. {1 x 3}
    5. 1
    6. 3
    """

    def __init__(
        self,
        space: TensorSpace,
    ):
        """
        Construct a _collection over which we will iterate later on.
        :param space: TensorSpace
        """
        self._position = 0
        self._substructure = space.substructure
        self._collection = []  # the list that is traversed during iterations, needs to be constructed
        element_list = [0]  # keeps track of the structure and is constructed during the initialization
        while len(element_list) > 0:
            collection_element = self._get_substructure(element_list)
            if collection_element:
                self._collection.append(collection_element)
                element_list.append(0)
            else:
                element_list = element_list[:-1]
                if element_list[-1] == 1:
                    # We have to find the last 0 element in the list and switch this to a one. The rest must be removed.
                    try:
                        index = element_list[::-1].index(0)
                    except ValueError:
                        # We reached the last element
                        break
                    element_list = element_list[:-index]
                element_list[-1] = 1

    def _get_substructure(
        self,
        element_list: list[int],
    ) -> Optional[TensorSpace]:
        """
        Auxiliary function for the iterator initialization.
        The idea is to loop through all subspaces of the space to construct the _collection list, which is later used
        in the iteration steps.

        :param element_list: A list of integers, which provides a map of the sub-spaces of the space.
        :return: Sub-Space or None
        """
        out_space = self._substructure
        for pos in element_list[:-1]:
            out_space = out_space[pos].substructure
        if out_space:
            return out_space[element_list[-1]]
        else:
            return None

    def __next__(
        self,
    ) -> TensorSpace:
        """
        Proceed to the next element.

        :return: The next-in-line TensorSpace element
        """
        try:
            value = self._collection[self._position]
            self._position += 1
        except IndexError:
            raise StopIteration()
        return value

    def get_next(
        self,
    ) -> Optional[TensorSpace]:
        """
        Same as the __next__ function but without raising an exception.

        :return:
        """
        try:
            value = self._collection[self._position]
            self._position += 1
        except IndexError:
            return None
        return value


class TensorSpace(Iterable):
    """
    An iterable class used to define the space of operators and states. Objects of this class can be compared and their
    subspace can be traversed with the SpaceIterator.
    """

    space_dict = {}  # Keep track which spaces already exist.

    def __init__(
        self,
        name: str,
        order: int,
        substructure: Optional[List[TensorSpace]] = None,
    ):
        self._name = name
        self._order = order
        self._substructure = substructure
        self._pure_space = not substructure
        if name not in self.space_dict.keys():
            self.space_dict[name] = self
        else:
            other = self.space_dict[name]
            if not (other.order == self.order and other.substructure == self.substructure):
                raise DuplicateSpaceError(f'A space with the name "{name}" and different properties already exists.')

    @property
    def name(
        self,
    ):
        return self._name

    @property
    def pure_space(
        self,
    ):
        return self._pure_space

    @property
    def order(
        self,
    ):
        return self._order

    @property
    def substructure(
        self,
    ):
        return self._substructure

    def __str__(
        self,
    ) -> str:
        return f"{self.name}-space: order = {self.order}"

    def __repr__(
        self,
    ) -> str:
        return self.__str__()

    def __add__(
        self,
        other: TensorSpace,
    ) -> TensorSpace:
        """
        Combine two spaces, if they are identical, return the space. Else, construct a new space, which contains the
        two spaces as a substructure

        :param other: TensorSpace
        :return: TensorSpace
        """
        if self == other:
            return self
        else:
            new_name = "{" + self.name + " x " + other.name + "}"
            new_order = min(self.order, other.order)
            return self.__class__(new_name, new_order, substructure=[self, other])

    def __eq__(
        self,
        other: TensorSpace,
    ) -> bool:
        """
        TensorSpaces cannot share a name if they have different properties. For equality, it is enough to check the name

        :param other: TensorSpace
        :return: True or False
        """
        if self.name == other.name:
            return True
        else:
            return False

    def __gt__(
        self,
        other: TensorSpace,
    ):
        """
        Order spaces according to their definition. The order is the first criteria. If there are coupled spaces, we
        further prefer shallower space to deep spaces. If those are the same we iterate through the spaces and repeat
        these criteria for each subspace. See also the SpaceIterator class. In case only one object has a substructure,
        we want this one on the first position.

        :param other: TensorSpace
        :return: True or False
        """
        if self.order > other.order:
            return True  # Those should be switched
        elif self.order == other.order:
            if self.substructure and other.substructure:
                if self.get_depth() < other.get_depth():
                    # both have identical order now the deepest substructure loses
                    return True
                elif self.get_depth() == other.get_depth():
                    self_iterator = SpaceIterator(self)
                    other_iterator = SpaceIterator(other)
                    next_self = self_iterator.get_next()
                    other_self = other_iterator.get_next()
                    while next_self and other_self:
                        if next_self.order > other_self.order:
                            return True
                        next_self = self_iterator.get_next()
                        other_self = other_iterator.get_next()
            elif self.substructure:
                return False  # Objects with substructure have higher rank by choice
            elif other.substructure:
                return True  # Objects with substructure have higher rank by choice
        return False

    def __ge__(
        self,
        other: TensorSpace,
    ):
        if self == other:
            return True
        elif self > other:
            return True
        return False

    def __lt__(
        self,
        other: TensorSpace,
    ):
        return not self >= other

    def __le__(
        self,
        other: TensorSpace,
    ):
        return not self > other

    def __ne__(
        self,
        other: TensorSpace,
    ):
        return not self == other

    def __iter__(
        self,
    ) -> SpaceIterator:
        return SpaceIterator(self)

    def get_depth(
        self,
    ) -> int:
        """
        Get the depth of the TensorSpace, by going recursively through the substructure.

        :return: Integer
        """
        if not self.substructure:
            return 0
        else:
            sub_structure_depth = max([sub_tensor.get_depth() for sub_tensor in self.substructure]) + 1
            return sub_structure_depth

    def contains(
        self,
        other: TensorSpace,
    ) -> bool:
        """
        Used to check if this TensorSpace contains the other space within it.

        :param other: TensorSpace
        :return: True or False
        """
        if self == other:
            return True
        if self.get_depth() == 0:
            return False
        for space in self:
            if space == other:
                return True
        return False

    def get_flat_basic_states(
        self,
    ) -> list[TensorSpace]:
        """
        Finds all basic TensorSpace objects (i.e., those without a substructure) and returns them as a list.
        :return: list of basic states
        """
        basic_states = []
        flat_basic_states = []
        if self.pure_space:
            basic_states.append(self)
        else:
            subspace_1, subspace_2 = self.substructure
            basic_states.extend(subspace_1.get_flat_basic_states())
            basic_states.extend(subspace_2.get_flat_basic_states())
        for bs in basic_states:
            if bs not in flat_basic_states:
                flat_basic_states.append(bs)  # Need to do it this way since the entries are not hashable
        if flat_basic_states:
            flat_basic_states.sort(key=lambda x: x.order)
        return flat_basic_states



from __future__ import annotations
from abc import ABC, abstractmethod
from sympy import Symbol
from typing import Optional, List

from stolpy.tensor_space import TensorSpace


class StateInterface(ABC):
    """
    Interface for BasicStates and CoupledStates
    """

    def __init__(
        self,
        angular_quantum_number: Symbol,
        space: TensorSpace,
        other_quantum_number: Optional[Symbol] = None,
        factor=1,
        substructure: Optional[List[StateInterface]] = None,
    ):
        self._angular_quantum_number = angular_quantum_number
        self._other_quantum_number = other_quantum_number
        self._factor = factor
        self._substructure = substructure
        self._space = space

    def __str__(
        self,
    ):
        if self.other_quantum_number:
            return f"|{self.other_quantum_number}{self.representation()}>"
        else:
            return f"|{self.representation()}>"

    def __repr__(
        self,
    ):
        return self.__str__()

    def __eq__(self, other: StateInterface) -> bool:
        if (
            self.angular_quantum_number == other.angular_quantum_number
            and self.other_quantum_number == other.other_quantum_number
            and self.factor == other.factor
            and self.substructure == other.substructure
        ):
            return True

    @property
    def angular_quantum_number(
        self,
    ):
        return self._angular_quantum_number

    @property
    def other_quantum_number(
        self,
    ):
        return self._other_quantum_number

    @property
    def substructure(
        self,
    ):
        return self._substructure

    @property
    def factor(
        self,
    ):
        return self._factor

    @property
    def space(
        self,
    ):
        return self._space

    @abstractmethod
    def representation(
        self,
    ) -> str:
        """
        Provide a string representation of each subclass. The representation is used in the __str__ function and can be
        appended for coupled states.
        :return: String
        """
        pass

    @abstractmethod
    def evaluate(
        self,
        symbolic_replace_dict: dict,
    ) -> str:
        """
        Evaluate all or some of the symbolic constituents of the state and return a string representation of it.
        :param symbolic_replace_dict: A dictionary that contains symbols and their replacement value as key-value pairs.
        :return: A String representation of the state, but with evaluated symbols
        """
        pass

    def couple(
        self,
        other: StateInterface,
        new_quantum_number: Symbol,
    ) -> CoupledState:
        """
        Couple another state to this state to form a new CoupledState object. During the coupling process, the states'
        constituents are ordered according to the order of their space object.
        :param other: Another state, either BasicState or CoupledState
        :param new_quantum_number: The symbol of the newly formed state
        :return: A CoupledState object
        """
        if self.space > other.space:
            first = other
            second = self
        else:
            first = self
            second = other
        new_factor = first.factor * second.factor
        new_space = first.space + second.space
        new_other_quantum_number = ""
        if first.other_quantum_number:
            new_other_quantum_number = first.other_quantum_number
        if second.other_quantum_number:
            new_other_quantum_number = Symbol(f"{new_other_quantum_number}{second.other_quantum_number}")
        if not new_other_quantum_number:
            new_other_quantum_number = None
        return CoupledState(
            angular_quantum_number=new_quantum_number,
            space=new_space,
            substructure=[first, second],
            other_quantum_number=new_other_quantum_number,
            factor=new_factor,
        )


class CoupledState(StateInterface):
    """
    CoupledState objects are a combination of two substates. They can in principle be initialized, but they can also be
    constructed using the couple function on StateInterface objects. Coupled states know their constituents, such that
    we can recouple them when we need it.
    """

    def __init__(
        self,
        angular_quantum_number: Symbol,
        space: TensorSpace,
        substructure: List[StateInterface],
        other_quantum_number: Optional[Symbol] = None,
        factor=1,
    ):
        super().__init__(angular_quantum_number, space, other_quantum_number, factor, substructure)

    def representation(
        self,
    ) -> str:
        return (
            f"{self.angular_quantum_number}"
            f"({self.substructure[0].representation()}{self.substructure[1].representation()})"
        )

    def evaluate(
        self,
        symbolic_replace_dict: dict,
    ) -> str:
        """
        A coupled state j has consists of two substates j1 and j2. The representation reads |j(j1j2)> if the substates
        are basic states or |j(j1(j1aj1b(...))j2(...)> if the substates are combined states as well. The substates are
        evaluated recursively.
        :param symbolic_replace_dict: A dictionary that contains symbols and their replacement value as key-value pairs.
        :return: A String representation of the state, but with evaluated symbols
        """
        return (
            f"{self.angular_quantum_number.subs(symbolic_replace_dict)}"
            f"({self.substructure[0].evaluate(symbolic_replace_dict)}"
            f"{self.substructure[1].evaluate(symbolic_replace_dict)})"
        )


class BasicState(StateInterface):
    """
    BasicState objects are states without a substructure. For us, those are the fundamental building blocks in our
    calculation. They mark the target to which we ultimately want to decouple our states and operators.
    """

    def __init__(
        self,
        angular_quantum_number: Symbol,
        space: TensorSpace,
        other_quantum_number: Optional[Symbol] = None,
        factor=1,
    ):
        super().__init__(angular_quantum_number, space, other_quantum_number, factor, substructure=None)

    def representation(
        self,
    ) -> str:
        return f"{self.angular_quantum_number}"

    def evaluate(
        self,
        symbolic_replace_dict: dict,
    ) -> str:
        """
        Replace all or some of the symbolic constituents by values according to the symbolic_replace_dict dictionary.
        :param symbolic_replace_dict: A dictionary that contains symbols and their replacement value as key-value pairs.
        :return: A String representation of the state, but with evaluated symbols
        """
        if self.other_quantum_number:
            other = self.other_quantum_number.subs(symbolic_replace_dict)
        else:
            other = ""
        return f"{other}{self.angular_quantum_number.subs(symbolic_replace_dict)}"



from __future__ import annotations
from typing import Optional
from sympy import sqrt, I

from stolpy.tensor_operator import TensorOperator


class WrongOperatorTypeError(Exception):
    """
    This error gets raised when the Input type of the TensorOperator is not correct in the TensorsFromVector class.
    """


class TensorFromVectors(object):
    """
    In equations, we are usually confronted with vectors rather than tensors. The transition from Vectors to Tensors
    is accompanied by specific factors, depending on the type of vector coupling. This class pretends that the
    basic tensors we construct are still vectors, and offers methods to join them with the vector operations
    scalar_product and vector_product. The result is the tensor representation with the correct factors.
    """

    @classmethod
    def scalar_product(
        cls,
        operator_1: TensorOperator,
        operator_2: TensorOperator,
    ) -> Optional[TensorOperator]:
        """
        Uses TensorOperator objects as input, pretends they are still vectors that are coupled via a scalar product
        and returns the tensor product
        {A_1 x B_1}_0 with the correct factor -sqrt(3).

        :param operator_1: TensorOperator
        :param operator_2: TensorOperator
        :return: TensorOperator or None, depending on the constellation
        """
        if not (isinstance(operator_1, TensorOperator) and isinstance(operator_2, TensorOperator)):
            raise WrongOperatorTypeError(f"Input type {type(operator_1)} with {type(operator_2)} is not allowed.")
        return operator_1.couple(operator_2, 0, -sqrt(3))

    @classmethod
    def vector_product(
        cls,
        operator_1: TensorOperator,
        operator_2: TensorOperator,
    ) -> Optional[TensorOperator]:
        """
        Uses TensorOperator objects as input, pretends they are still vectors that are coupled via a vector product
        and returns the tensor product
        {A_1 x B_1}_1 with the correct factor -I sqrt(2).

        :param operator_1: TensorOperator
        :param operator_2: TensorOperator
        :return: TensorOperator or None, depending on the constellation
        """
        if not (isinstance(operator_1, TensorOperator) and isinstance(operator_2, TensorOperator)):
            raise WrongOperatorTypeError(f"Input type {type(operator_1)} with {type(operator_2)} is not allowed.")
        return operator_1.couple(operator_2, 1, -I * sqrt(2))



from __future__ import annotations
import copy
from abc import ABC, abstractmethod
from typing import Union
from sympy.physics.wigner import wigner_6j, wigner_9j
from sympy import Symbol


def factor_eval(
    factor: Union[int, float, Symbol],
    symbol_replace_dict: dict,
):
    """
    Auxiliary function to evaluate the factor. The function checks the type and substitutes potential symbols.
    :param factor: The factor from any of the objects below.
    :param symbol_replace_dict: A dictionary that contains any unresolved symbols and their replacement values as key-
           value pairs.
    :return: The evaluated form of the factor.
    """
    try:
        return factor.subs(symbol_replace_dict)
    except AttributeError:
        return factor


class SymbolicWigner(ABC):
    """
    An auxiliary construct that helps to hold the evaluation of wigner_6j and wigner_9j symbols.
    """

    def __init__(
        self,
    ):
        self.factor = None

    @abstractmethod
    def evaluate(
        self,
        symbol_replace_dict: dict,
    ):
        """
        Perform a substitution of the symbols in the object according to the input dictionary and evaluate the symbolic
        6j and or 9j symbols.

        :param symbol_replace_dict: A dictionary that contains symbols and their replacement value as key-value pairs.
        :return: the evaluated result.
        """
        pass

    @abstractmethod
    def __str__(
        self,
    ):
        """
        Print the object's components in a symbolic way, e.g. SixJ(j1 j2 j3; j4 j5 j6).
        :return: String representation of the object.
        """
        pass

    def __repr__(
        self,
    ):
        return self.__str__()

    def __mul__(
        self,
        other: Union[SymbolicWigner, int, float, Symbol],
    ) -> SymbolicWigner:
        """
        Multiplication with another object. The other object needs to be of type
            - SymbolicWigner -> Return Value is SymbolicWignerComposite
            - int, float, Symbol -> Changes only the factor and returns a Symbolic6j/9j object
        :param other: SymbolicWigner, int, float, or Symbol type
        :return: Either SymbolicWignerComposite (other is of type SymbolicWigner) or Symbolic6j/9j (else)
        """
        if isinstance(other, SymbolicWigner):
            return SymbolicWignerComposite(self, other)
        else:
            new_self = copy.deepcopy(self)
            new_self.factor *= other
            return new_self

    def __rmul__(
        self,
        other,
    ):
        return self.__mul__(other)

    def __truediv__(
        self,
        other,
    ):
        return self.__mul__(1 / other)


class SymbolicWignerComposite(SymbolicWigner):
    """
    Stores a collection of SymbolicWigner symbols. This allows to multiply 6j and 9j symbols and store them in a
    symbolic way.
    """

    def __init__(
        self,
        instance_1,
        instance_2,
    ):
        super().__init__()
        self.children = []
        self.add_children(instance_1)
        self.add_children(instance_2)
        self.factor = 1

    def __str__(
        self,
    ):
        return f"{self.factor} * (" + " * ".join([str(child) for child in self.children]) + ")"

    def __mul__(
        self,
        other: Union[SymbolicWigner, int, float, Symbol],
    ) -> SymbolicWignerComposite:
        """
        Overrides the __mul__ function in the interface. In case the "other" input is of type SymbolicWigner
        :param other: SymbolicWigner object. Depending on the exact subtype, the add_children class extends the
                      children list or appends the object.
        :return: SymbolicWignerComposite with updated children or factor.
        """
        new_self = copy.deepcopy(self)
        if isinstance(other, SymbolicWigner):
            new_self.factor *= other.factor
            new_self.add_children(other)
        else:
            new_self.factor *= other
        return new_self

    def evaluate(
        self,
        symbol_replace_dict: dict,
    ):
        """
        Replace all symbols by numbers. If not all symbols. See also Symbolic6j and Symbolic9j. An error is raised in
        case the input of the 6j and/or 9j symbols in this collection are not completely replaced.

        :param symbol_replace_dict: dictionary that contains symbols as keys and their replacement values as values
        :return: the evaluated collection of 6j and/or 9j symbols.
        """
        sub_symbol_value = 1
        for child in self.children:
            sub_symbol_value *= child.evaluate(symbol_replace_dict)
        factor = factor_eval(self.factor, symbol_replace_dict)
        return factor * sub_symbol_value

    def add_children(
        self,
        other: SymbolicWigner,
    ) -> None:
        """
        Extends children in case other is of same type, otherwise append other object to children. The action happens
        in place.
        :param other: SymbolicWigner object
        :return: None
        """
        if isinstance(other, self.__class__):
            self.children.extend(other.children)
        else:
            self.children.append(other)


class Symbolic6j(SymbolicWigner):
    """
    Auxiliary construct to handle Wigner-6j symbols in a symbolic way.
    """

    six_j_cache = {}

    def __init__(
        self,
        j1,
        j2,
        j3,
        j4,
        j5,
        j6,
        factor: Union[int, float, Symbol] = 1,
    ):
        super().__init__()
        self.j1 = j1
        self.j2 = j2
        self.j3 = j3
        self.j4 = j4
        self.j5 = j5
        self.j6 = j6
        self.factor = factor

    def __str__(self):
        return f"{self.factor} * SixJ({self.j1} {self.j2} {self.j3}; {self.j4} {self.j5} {self.j6})"

    def evaluate(
        self,
        symbol_replace_dict: dict,
    ):
        """
        Replace all symbols by numbers. If not all symbols in the 6j symbol are provided, the wigner_6j function will
        raise an error.

        :param symbol_replace_dict: dictionary that contains symbols as keys and their replacement values as values
        :return: the evaluated 6j symbol
        """
        j1, j2, j3, j4, j5, j6 = [
            x.subs(symbol_replace_dict) if isinstance(x, Symbol) else x
            for x in (
                self.j1,
                self.j2,
                self.j3,
                self.j4,
                self.j5,
                self.j6,
            )
        ]
        factor = factor_eval(self.factor, symbol_replace_dict)
        key = (j1, j2, j3, j4, j5, j6)
        if key in self.six_j_cache:
            six_j_value = self.six_j_cache[key]
        else:
            six_j_value = wigner_6j(j1, j2, j3, j4, j5, j6)
            self.six_j_cache[key] = six_j_value
        return factor * six_j_value


class Symbolic9j(SymbolicWigner):
    """
    Auxiliary construct to handle Wigner-9j symbols in a symbolic way.
    """

    nine_j_cache = {}

    def __init__(
        self,
        j1,
        j2,
        j3,
        j4,
        j5,
        j6,
        j7,
        j8,
        j9,
        factor: Union[int, float, Symbol] = 1,
    ):
        super().__init__()
        self.j1 = j1
        self.j2 = j2
        self.j3 = j3
        self.j4 = j4
        self.j5 = j5
        self.j6 = j6
        self.j7 = j7
        self.j8 = j8
        self.j9 = j9
        self.factor = factor

    def __str__(
        self,
    ):
        return (
            f"{self.factor} * NineJ({self.j1} {self.j2} {self.j3}; {self.j4} {self.j5} {self.j6}; {self.j7} "
            + f"{self.j8} {self.j9})"
        )

    def evaluate(
        self,
        symbol_replace_dict: dict,
    ):
        """
        Replace all symbols by numbers. If not all symbols in the 9j symbol are provided, the wigner_9j function will
        raise an error.

        :param symbol_replace_dict: dictionary that contains symbols as keys and their replacement values as values
        :return: the evaluated 9j symbol
        """
        j1, j2, j3, j4, j5, j6, j7, j8, j9 = [
            x.subs(symbol_replace_dict) if isinstance(x, Symbol) else x
            for x in (self.j1, self.j2, self.j3, self.j4, self.j5, self.j6, self.j7, self.j8, self.j9)
        ]
        factor = factor_eval(self.factor, symbol_replace_dict)
        key = (j1, j2, j3, j4, j5, j6, j7, j8, j9)
        if key in self.nine_j_cache:
            nine_j_value = self.nine_j_cache[key]
        else:
            nine_j_value = wigner_9j(j1, j2, j3, j4, j5, j6, j7, j8, j9)
            self.nine_j_cache[key] = nine_j_value
        return factor * nine_j_value



from __future__ import annotations
from sympy import KroneckerDelta, Symbol
from typing import Union, Optional, List
from abc import ABC, abstractmethod
import copy
import logging

from stolpy.tensor_algebra import jsc
from stolpy.quantum_states import StateInterface
from stolpy.tensor_algebra import TensorAlgebra
from stolpy.tensor_operator import TensorOperator, TensorOperatorComposite, TensorOperatorInterface
from stolpy.symbolic_wigner import Symbolic6j, Symbolic9j, SymbolicWigner, factor_eval

logging.basicConfig(level=logging.WARNING)


class StateMismatchError(Exception):
    """Raised when the space of the final state does not match that of the initial one or if the operator acts on
    different spaces than the states provide."""

    pass


class RankMismatchError(Exception):
    """Raised when the rank of operators does not match."""

    pass


class MatrixElementInterface(ABC):
    """
    Interface for Leaf and Composite MatrixElement classes.
    """

    @abstractmethod
    def decouple(
        self,
    ):
        """
        Decouple tensor operators and the tensor space. The result is a reduced matrix element of operators, i.e.,
        a product of operators that act in space x encompassed by the state for this space. E.g.
        <j'(j1'j2')| {O1 x O2} |j(j1j2)> -> <j1'||O1||j1> * <j2'||O2||j2> (Note the example is simplified, it does not
        consider ranks of the tensor operators). It can also return a sum of reduced matrix elements. Some Reduced
        Matrix Elements can be decoupled further.
        :return: ReducedMatrixElement
        """
        pass

    @abstractmethod
    def __str__(
        self,
    ) -> str:
        """
        Return a string representation of Matrix Elements.
        :return: String
        """
        pass

    @abstractmethod
    def evaluate(
        self,
        symbolic_replace_dict: dict,
    ):
        """
        Evaluate the matrix elements by replacing all or some of the symbols in the objects.
        :param symbolic_replace_dict: A dictionary that contains symbols and their replacement value as key-value pairs.
        :return: Evaluation of the expression.
        """
        pass


class ReducedMatrixElementComposite(MatrixElementInterface):
    """
    A collection of ReducedMatrixElement objects. The class contains a list for the reduced_matrix_element_a and
    reduced_matrix_element_b (short rme_a/b). The lists are ordered and contain a MatrixElementInterface list.
    Elements between list a and b with the same index are multiplied, while this produced is added between different
    indexes, e.g., rme_a = [me_a1, me_a2, me_a3, ...], rme_b = [me_b1, me_b2, me_b3, ...] means the actual operator
    structure is me_a1 * me_b1 + me_a2 * me_b2 + me_a3 * me_b3 + ...
    Note that the lists have the same length by construction. The factor list works similarly.

    While it is possible to initialize this element directly, we recommend going via MatrixElement objects and use
    their decouple function.
    """

    def __init__(
        self,
        reduced_me_a: Union[List[ReducedMatrixElement], ReducedMatrixElement],
        reduced_me_b: Union[List[ReducedMatrixElement], ReducedMatrixElement],
        factor,
    ):
        if isinstance(reduced_me_a, list):
            self._reduced_me_a = reduced_me_a
            self._reduced_me_b = reduced_me_b
            self._factor = factor
        else:
            self._reduced_me_a = [reduced_me_a]
            self._reduced_me_b = [reduced_me_b]
            self._factor = [factor]

    @property
    def reduced_matrix_element_a(
        self,
    ):
        return self._reduced_me_a

    @property
    def reduced_matrix_element_b(
        self,
    ):
        return self._reduced_me_b

    @property
    def factor(
        self,
    ):
        return self._factor

    @property
    def children(
        self,
    ):
        """
        :return: iterable object that contains the factor as the first element, the reduced_matrix_element_a as the
        second and the reduced_matrix_element_b as the last.
        """
        return zip(self.factor, self.reduced_matrix_element_a, self.reduced_matrix_element_b)

    def __str__(
        self,
    ):
        content = [f"{x[0]} * {x[1]}{x[2]}" for x in self.children]
        return " + ".join(content)

    def __eq__(
        self,
        other: ReducedMatrixElementComposite,
    ) -> bool:
        if self.children == other.children:
            return True
        else:
            return False

    def __neg__(
        self,
        other: ReducedMatrixElementComposite,
    ) -> bool:
        return not self.__eq__(other)

    def _full_composite_decouple(
        self,
    ) -> ReducedMatrixElementComposite:
        """
        Fully decouples a ReducedMatrixElementComposite if possible. The decoupling is done inplace. The object is
        returned afterwards. The decoupling happens recursively, thus, it is capable to handle arbitrary structures.
        :return: ReducedMatrixElementComposite
        """
        changed = True
        while changed:
            changed = False
            for _, me_a, me_b in self.children:
                me_a_copy = copy.deepcopy(me_a)
                me_b_copy = copy.deepcopy(me_b)
                me_a.decouple()
                me_b.decouple()
                if me_a != me_a_copy or me_b != me_b_copy:
                    changed = True
        return self

    def decouple(
        self,
    ) -> ReducedMatrixElementComposite:
        """
        This class already contains matrix elements that have been decoupled at least once. In some situations,
        it is possible to decouple the operators further. E.g., the structure
        <j'(j_ab'(j_a'j_b')j_c')| {{A x B} x C} |j(j_ab(j_aj_b)j_c)>
        decoupled once reads
        <j_ab'(j_a'j_b')|| {A x B} ||j_ab(j_aj_b)> <j_c'|| C ||j_c>.
        This can be decoupled again to
        <j_a'|| A ||j_a> <j_b')|| B ||j_b> <j_c'|| C ||j_c>.
        :return: ReducedMatrixElementComposite
        """
        new_me_a = []
        new_me_b = []
        for _, me_a, me_b in self.children:
            new_composite_a = me_a.decouple()
            new_composite_b = me_b.decouple()

            if new_composite_a:
                new_me_a.append(new_composite_a._full_composite_decouple())
            else:
                new_me_a.append(me_a)

            if new_composite_b:
                new_me_b.append(new_composite_b._full_composite_decouple())
            else:
                new_me_b.append(me_b)

        self._reduced_me_a = new_me_a
        self._reduced_me_b = new_me_b
        return self

    def append(
        self,
        other: ReducedMatrixElementComposite,
    ) -> None:
        """
        Append another ReduceMatrixElementComposite to the existing one.
        :param other: ReducedMatrixElementComposite object.
        :return: This action happens in place and does not return anything
        """
        self._reduced_me_a.extend(other.reduced_matrix_element_a)
        self._reduced_me_b.extend(other.reduced_matrix_element_b)
        self._factor.extend(other.factor)

    def me_evaluate(
        self,
        symbolic_replace_dict: dict,
    ):
        """
        Auxiliary function made so that the same structure as MatrixElements and ReducedMatrixElements is met.
        This function just uses the evaluate function.
        :param symbolic_replace_dict:
        :return: The evaluated expression where symbolic components are fully or partially replaced by values.
        """
        return self.evaluate(symbolic_replace_dict)

    def evaluate(
        self,
        symbolic_replace_dict: dict,
    ):
        """
        Recursively go through all children and call their evaluate function to replace some or all of the symbolic
        contributions in the expression.
        :param symbolic_replace_dict: A dictionary that contains symbols and their replacement value as key-value pairs.
        :return: The evaluated expression where symbolic components are fully or partially replaced by values.
        """
        self.decouple()
        ret_val = None
        for factor, me_a, me_b in self.children:
            if isinstance(factor, SymbolicWigner):
                term = factor.evaluate(symbolic_replace_dict)
            else:
                term = factor_eval(factor, symbolic_replace_dict)
            term *= me_a.me_evaluate(symbolic_replace_dict)
            term *= me_b.me_evaluate(symbolic_replace_dict)
            if ret_val:
                ret_val += term
            else:
                ret_val = term
        return ret_val


class BasicMatrixElementLeafInterface(MatrixElementInterface):
    """
    Interface for ReducedMatrixElement and MatrixElement objects.
    """

    def __init__(
        self,
        bra_state: StateInterface,
        ket_state: StateInterface,
        operator: Union[TensorOperator, TensorOperatorComposite, None],
        factor: Union[int, float, Symbol] = 1,
        recouple=True,
    ):
        self._bra = bra_state
        self._ket = ket_state

        if bra_state.space != ket_state.space:
            raise StateMismatchError("[ERROR] Final and initial states do not have the same space configuration.")
        state_basic_spaces = bra_state.space.get_flat_basic_states()
        operator_basic_spaces = operator.space.get_flat_basic_states()
        if state_basic_spaces != operator_basic_spaces:
            logging.debug("Operator spaces do not match State spaces")
            # Case A: The operator does not contain contributions for all spaces in the state -> Add unit operators
            for sbs in state_basic_spaces:
                if sbs not in operator_basic_spaces:
                    unit_operator = TensorOperator(rank=0, factor=1, space=sbs, symbol=Symbol(f"I_{sbs.name}"))
                    operator = operator.couple(unit_operator, operator.rank, 1)
                    logging.debug(f"Adding the unit operator {unit_operator}")
            # Case B: The operator acts on spaces which are not present in the state -> Throw an error
            for obs in operator_basic_spaces:
                if obs not in state_basic_spaces:
                    raise StateMismatchError("[ERROR] The operator acts on spaces that are not present in the states.")

        if recouple:
            self._operator = TensorAlgebra.recouple(operator)  # Simplifies the operator structure
        else:
            self._operator = operator
        if operator:
            self._factor = factor * operator.factor
        else:
            self._factor = factor

    @property
    def bra(
        self,
    ):
        return self._bra

    @property
    def ket(
        self,
    ):
        return self._ket

    @property
    def operator(
        self,
    ):
        return self._operator

    @property
    def factor(
        self,
    ):
        return self._factor

    def __str__(
        self,
    ) -> str:
        """
        Visual representation of the object. See _state_representation for more information.

        :return: String
        """
        if isinstance(self.operator, TensorOperatorComposite):
            return " + ".join(
                [
                    f"{op.factor} * {self._state_representation(op)}"
                    if op.factor != 1
                    else f"{self._state_representation(op)}"
                    for op in self.operator.children
                ]
            )
        else:
            factor = self.operator.factor
            if factor != 1:
                output = f"{factor} * {self._state_representation(self.operator)}"
            else:
                output = f"{self._state_representation(self.operator)}"
            return output

    def __repr__(
        self,
    ):
        return self.__str__()

    @abstractmethod
    def _state_representation(
        self,
        operator: TensorOperatorInterface,
    ) -> str:
        """
        Auxiliary function that is used in the __str__ function. The function is implemented differently between
        MatrixElement objects and ReducedMatrixElement objects. It also takes care of a bra and ket notation, as the
        default implementation in State objects is the ket notation. The state representation takes an operator as
        input to keep it more flexible.

        :param operator: TensorOperatorInterface object
        :return: A representative String.
        """
        pass

    @abstractmethod
    def _basic_decouple(
        self,
        operator: TensorOperatorInterface,
    ):
        """
        Auxiliary decouple formalism that helps to distinguish between MatrixElements and ReducedMatrixElements. It
        also gives more flexibility, as it will decouple an arbitrary (part of the) operator.

        :param operator: TensorOperatorInterface object
        :return: ReducedMatrixElementComposite object.
        """
        pass

    def decouple(
        self,
    ) -> Optional[ReducedMatrixElementComposite]:
        """
        Decouples the given matrix element one level. A different treatment happens for MatrixElement objects and
        ReducedMatrixElement objects via the _basic_decouple function. It is possible that the result can be recoupled
        further. Use the function full_decouple to get the final expression
        """
        if isinstance(self.operator, TensorOperatorComposite):
            red_me_composite = ReducedMatrixElementComposite([], [], [])
            for op in self.operator.children:
                additional_red_me_composite = self._basic_decouple(op)
                if additional_red_me_composite:
                    red_me_composite.append(additional_red_me_composite)
            return red_me_composite
        else:
            return self._basic_decouple(self.operator)

    def full_decouple(
        self,
    ) -> Optional[ReducedMatrixElementComposite]:
        """
        Decouples the given matrix element. A different treatment happens for MatrixElement objects and
        ReducedMatrixElement objects via the _basic_decouple function.
        """
        composite = self.decouple()
        return composite.decouple()  # Make use of the full decoupling routine in composites

    def me_evaluate(
        self,
        symbol_replace_dict: dict,
    ):
        """
        Replaces all or some of the symbolic expressions with values and returns the result. This function is called
        within the Composite evaluation routine.

        :param symbol_replace_dict: A dictionary that contains symbols and their replacement value as key-value pairs.
        :return: The evaluated expression where symbolic components are fully or partially replaced by values.
        """
        if isinstance(self, ReducedMatrixElement):
            separator = "||"
        else:
            separator = "|"
        me = Symbol(
            f"<{self.bra.evaluate(symbol_replace_dict)}{separator}{self.operator.to_expression_no_factor()}"
            f"{separator}{self.ket.evaluate(symbol_replace_dict)}>"
        )
        return me

    def evaluate(
        self,
        symbol_replace_dict: dict,
    ):
        composite = self.full_decouple()
        return composite.evaluate(symbol_replace_dict)


class MatrixElement(BasicMatrixElementLeafInterface):
    """
    Allows the user to create a matrix element object, containing a bra state, ket state and an operator.
    We recommend to use this class to define objects and create other class instances by using the recouple function.
    """

    def _state_representation(
        self,
        operator: TensorOperatorInterface,
    ) -> str:
        """
        Auxiliary function that is used in the __str__ function. For completeness, we also provide the projection of the
        angular quantum number. It is, however, not used in the evaluation process in this work. One needs it during the
        application of the Wigner-Eckart theorem.

        :param operator: TensorOperatorInterface object
        :return: A representative String.
        """
        ket_angular_quantum_projection = f"m_{self.ket.angular_quantum_number}"
        bra_angular_quantum_projection = f"m_{self.bra.angular_quantum_number}"
        return (
            f"<{str(self.bra)[1:-1]}{bra_angular_quantum_projection}|{operator.to_expression_no_factor()}"
            f"{str(self.ket)[:-1]}" + f"{ket_angular_quantum_projection}>"
        )

    def _basic_decouple(
        self,
        operator: TensorOperator,
    ) -> Optional[ReducedMatrixElementComposite]:
        """
        The function is called for TensorOperators. It decouples potential parts of the complete self.operator.

        :param operator: A part of the self.Operator. It is broken down to TensorOperator objects.
        :return: ReducedMatrixElementComposite or None
        """
        if not self.bra.substructure or not self.ket.substructure:
            logging.info("There is nothing to decouple")
            return None

        tensor_a, tensor_b = operator.substructure
        if tensor_a.rank != tensor_b.rank:
            raise RankMismatchError("[ERROR] Ranks of sub-operators must match")

        rank = tensor_a.rank

        bra_a, bra_b = self.bra.substructure
        ket_a, ket_b = self.ket.substructure
        bra_j = self.bra.angular_quantum_number
        ket_j = self.ket.angular_quantum_number
        bra_j_a = bra_a.angular_quantum_number
        bra_j_b = bra_b.angular_quantum_number
        ket_j_a = ket_a.angular_quantum_number
        ket_j_b = ket_b.angular_quantum_number

        factor = (
            operator.factor
            * pow(-1, bra_j + bra_j_b + ket_j_a + rank)
            / jsc(rank)
            * KroneckerDelta(bra_j, ket_j)
            * Symbolic6j(bra_j_a, bra_j_b, bra_j, ket_j_b, ket_j_a, rank)
        )
        reduced_matrix_element_a = ReducedMatrixElement(bra_a, ket_a, tensor_a)
        reduced_matrix_element_b = ReducedMatrixElement(bra_b, ket_b, tensor_b)
        return ReducedMatrixElementComposite(reduced_matrix_element_a, reduced_matrix_element_b, factor)


class ReducedMatrixElement(
    BasicMatrixElementLeafInterface,
):
    """
    A class for reduced matrix elements. While it is possible, we do not recommend to initialize objects of this class
    directly. We recommend users to initialize MatrixElement objects and use the decouple function to obtain reduced
    matrix elements. Reduced matrix elements can be calculated using the Wigner-Eckart theorem.
    """

    def __init__(
        self,
        bra_state: StateInterface,
        ket_state: StateInterface,
        operator: Union[TensorOperator, TensorOperatorComposite],
    ):
        super().__init__(bra_state, ket_state, operator, factor=1, recouple=False)
        self._value = None

    def __eq__(
        self,
        other: ReducedMatrixElement,
    ) -> bool:
        if self.bra == other.bra and self.ket == other.ket and self.operator == other.operator:
            return True
        else:
            return False

    def __neg__(
        self,
        other: ReducedMatrixElement,
    ) -> bool:
        return not self.__eq__(other)

    def _state_representation(
        self,
        operator,
    ) -> str:
        """
        Returns a string representation of the matrix element for a given operator.

        :param operator: The operator or sub-operator of the object
        :return: String representation
        """
        return f"<{str(self.bra)[1:-1]}||{operator.to_expression_no_factor()}|{self.ket}"

    def _basic_decouple(
        self,
        operator: TensorOperator,
    ) -> Optional[ReducedMatrixElementComposite]:
        """
        In some cases, it is possible to decouple reduced matrix elements even further (e.g., when there are three
        operators from distinct spaces). This method decouples those cases or returns None if no further recoupling is
        possible.

        :param operator: TensorOperator object
        :return: ReducedMatrixElementComposite or None
        """
        if operator.substructure:
            tensor_a, tensor_b = operator.substructure
            if tensor_a.space != tensor_b.space:
                rank = operator.rank
                a = tensor_a.rank
                b = tensor_b.rank
                bra_a, bra_b = self.bra.substructure
                ket_a, ket_b = self.ket.substructure
                bra_j = self.bra.angular_quantum_number
                ket_j = self.ket.angular_quantum_number
                bra_j_a = bra_a.angular_quantum_number
                bra_j_b = bra_b.angular_quantum_number
                ket_j_a = ket_a.angular_quantum_number
                ket_j_b = ket_b.angular_quantum_number
                new_factor = jsc(bra_j, ket_j, rank) * Symbolic9j(
                    a, b, rank, bra_j_a, bra_j_b, bra_j, ket_j_a, ket_j_b, ket_j
                )
                reduced_matrix_element_a = ReducedMatrixElement(bra_a, ket_a, tensor_a)
                reduced_matrix_element_b = ReducedMatrixElement(bra_b, ket_b, tensor_b)
                return ReducedMatrixElementComposite(
                    reduced_matrix_element_a, reduced_matrix_element_b, new_factor * self.factor
                )
        logging.info(f"Further decoupling not possible for reduced matrix element {self}")
        return None
